;;;; levenshtein-sequence-vector.scm -*- Scheme -*-
;;;; Kon Lovett, Mar '20
;;;; Kon Lovett, Apr '12
;;;; Kon Lovett, May '06

(include-relative "levenshtein-sequence-interface")

(module levenshtein-sequence-vector SEQUENCE-OPER

(import scheme
  (chicken base)
  (chicken type)
  (chicken fixnum)
  vector-lib
  (only record-variants define-record-type-variant)
  (only type-errors error-argument-type))

;;;

(include-relative "levenshtein-sequence-vector.types")

;;;

(define shared-vector 'shared-vector)
(define-record-type-variant shared-vector (unchecked inline unsafe)
  (%make-shared-vector vec start end)
  %shared-vector?
  (vec %shared-vector-vector)
  (start %shared-vector-start)
  (end %shared-vector-end))

;;;

(define (*vector-resolve vec start end)
  (assert (and (fixnum? start) (fixnum? end)))
  (let loop ((vec vec) (start start) (end end))
    (if (%shared-vector? vec)
        (*vector-resolve (%shared-vector-vector vec)
                         (fx+ (%shared-vector-start vec) start)
                         (fx+ (%shared-vector-start vec) (fx- end start)))
        (values vec start end)) ) )

(define (*vector? obj)
  (or (vector? obj) (%shared-vector? obj)) )

(define (error/type/sequence-vector loc obj #!optional nam)
  (error-argument-type loc obj "not vector or shared-vector" nam) )

(define (check-*vector loc obj)
  (unless (*vector? obj) (error/type/sequence-vector loc obj))
  obj )

(define (*vector-length vec)
  (cond ((vector? vec)
          (vector-length vec))
        ((%shared-vector? vec)
          (fx- (%shared-vector-end vec) (%shared-vector-start vec)))
        (else
          (error/type/sequence-vector '*vector-length vec)) ) )

(define (*vector-ref vec idx)
  (assert (and (fixnum? idx) (or (zero? idx) (positive? idx))))
  (cond ((vector? vec)
          (vector-ref vec idx))
        ((%shared-vector? vec)
          (let ((ridx (fx+ (%shared-vector-start vec) idx)))
            (assert (fx< ridx (%shared-vector-end vec)) 'vector-ref "out of range" idx)
            (*vector-ref (%shared-vector-vector vec) ridx)))
        (else
          (error/type/sequence-vector '*vector-ref vec)) ) )

(define (*vector-for-each f vec . vectors)
  (define (vec-for-each f vec start end)
    (do ((i start (fx+ i 1)))
        ((fx<= end i))
      (f i (*vector-ref vec i)) ) )
  (if (null? vectors)
      (vec-for-each f vec 0 (*vector-length vec))
      (error 'vector-for-each "multiple vector support not implemented") ) )

(define (subvector/shared vec start #!optional (end (*vector-length vec)))
  (assert (and (fixnum? start) (fixnum? end) (fx< start end)))
  (if (*vector? vec)
      (let-values (((vec start end) (*vector-resolve vec start end)))
         (%make-shared-vector vec start end))
      (error/type/sequence-vector 'subvector/shared vec) ) )

(define (*vector-prefix-length elm-eql v1 v2
          #!optional (s1 0) (e1 (*vector-length v1)) (s2 0) (e2 (*vector-length v2)))
  (assert (and (fixnum? s1) (fixnum? s2) (fixnum? e1) (fixnum? e2) (fx< s1 e1) (fx< s2 e2)))
  (let loop ((i1 s1) (i2 s2))
    (cond ((or (fx>= i1 e1) (fx>= i2 e2))
            (if (fx> i1 e1) 0 (fx- i1 s1)))
          ((elm-eql (*vector-ref v1 i1) (*vector-ref v2 i2))
            (loop (fx+ i1 1) (fx+ i2 1)))
          (else
            (fx- i1 s1)))) )

(define (*vector-suffix-length elm-eql v1 v2
          #!optional (s1 0) (e1 (*vector-length v1)) (s2 0) (e2 (*vector-length v2)))
  (assert (and (fixnum? s1) (fixnum? s2) (fixnum? e1) (fixnum? e2) (fx< s1 e1) (fx< s2 e2)))
  (let loop ((i1 (fx- e1 1)) (i2 (fx- e2 1)))
    (cond ((or (fx<= i1 s1) (fx<= i2 s2))
            (if (fx< i1 s1) 0 (fx- e1 (fx+ i1 1))))
          ((elm-eql (*vector-ref v1 i1) (*vector-ref v2 i2))
            (loop (fx- i1 1) (fx- i2 1)))
          (else
            (fx- e1 (fx+ i1 1))))) )

;;;

(define check-sequence check-*vector)
(define sequence-length *vector-length)
(define sequence-prefix-length *vector-prefix-length)
(define sequence-suffix-length *vector-suffix-length)
(define sequence-for-each *vector-for-each)
(define subsequence/shared subvector/shared)

) ;module levenshtein-sequence-vector
